"""General facilities for generating reports.

.. sidebar:: Contents

    .. contents::
        :local:
        :depth: 1

To do scientific research in terms of reproducibility and traceability it's
highly necessary to report all the steps done on a given dataset and never
separate the dataset from its metadata. Similarly, a recipe in context
of recipe-driven data analysis stores a lot of relevant information on what
tasks have been performed on a series of datasets (and even more so the
recipe history). However, having a dataset (or recipe history) containing
all these metadata is only useful if there are easy ways to retrieve and
present the information stored. This is the task of reports. This module
provides functionality to create reports based on templates provided either
by the user or by the package as such.


"Batteries included": Templates contained in the package
========================================================

The "batteries included" approach of Python itself is probably responsible
to a great deal for the success of Python as a language. ASpecD, similarly,
tries to provide you with a sensible set of tools you need for your routine
data analysis in spectroscopy. Reports are no exception to that rule.

Hence, ASpecD comes bundled with a (growing) series of templates allowing
you to create reports of datasets and alike. Thus, getting access to all
information stored in a single dataset is as simple as calling a single
reporter, and in context of recipe-driven data analysis, it is even simpler:

.. code-block:: yaml

    - kind: report
      type: LaTeXReporter
      properties:
        template: dataset.tex
        filename: report.tex
      compile: true

This would create a report of a dataset that is then stored in the file
``report.tex``, using the template ``dataset.tex`` bundled with the ASpecD
package. As you even set ``compile`` to true, it would even compile the
LaTeX report, including all figures generated during cooking the recipe and
referenced from within the report. Hence, you end up in your current
directory with both, a LaTeX file ``report.tex`` and a PDF file ``report.pdf``.

But what if you don't like the way the bundled templates look like? Don't
worry, we've got you covered: Simply provide a relative or absolute path to
your own template, even with the same name. Hence, in the above example,
if you place a file ``dataset.tex`` in the directory you serve the recipe
from, it will be used instead of the bundled one. Developers of other
packages can do the same and provide templates with the same name as those
provided by the ASpecD package. Theirs will take precedence. For details,
see below.


Output formats
==============

Template engines provide means of a "separation of concerns" (a term coined
by Edsger W. Dijkstra and highly important not only in software
development): The data source is entirely independent of the formatting of
the report, and one and the same template engine can be used to create
reports in a multitude of output formats.

Currently, the ASpecD framework supports two output formats (more may be
added in the future): plain text (txt) and LaTeX. The respective reporter
classes are:

  * :class:`TxtReporter`
  * :class:`LaTeXReporter`

The bundled templates used with the respective reporters are stored within
the ``templates/report/`` directory of the ASpecD package, and here within
subdirectories (``txt``, ``latex``) for each of the formats. This makes it
easy to add additional formats and to shorten the template paths when using
the reporters.


Choosing the language of a report
=================================

While ASpecD is currently designed to predominantly use English as its
native language, reports support different languages, provided the templates
for the respective language are contained in the package.

To select a language (different than the default, English), set the
:attr:`Reporter.language` attribute accordingly. By default, two-letter ISO
codes are used, such as ``en`` (English) and ``de`` (German). Please note
that this setting will only have an effect if templates for the requested
language are provided either by the ASpecD package or by the package based
on ASpecD.

In context of recipe-driven data analysis, getting a report on a dataset in
German rather than in English, use the following:


.. code-block:: yaml

    - kind: report
      type: LaTeXReporter
      properties:
        language: de
        template: datensatz.tex
        filename: bericht.tex
      compile: true


Note that the name of the template changes as well, as it would not make too
much sense to have English filenames for templates written in other
languages.

.. important::

    As the ASpecD framework natively uses the English language for
    descriptions, parameter names and alike, providing templates in
    different languages will likely result in a mixture of languages within
    the final report.


Package developers: Organisation of templates
=============================================

As mentioned above, the bundled templates used with the respective reporters
are stored within the ``templates/report/`` directory of the ASpecD package.
Each output format has its own subdirectory, currently existing directories
are ``txt`` and ``latex``. Furthermore, to allow for different languages,
each of these directories can (and should) contain another interlayer in
terms of directories representing the languages. By default, two-letter ISO
codes are used, such as ``en`` (English) and ``de`` (German).

Currently, the ASpecD template organisation looks similar to the following:

.. code-block::

    templates/
        report/
            latex/
                de/
                    basis.tex
                    datensatz.tex
                    ...
                en/
                    base.tex
                    dataset.tex
                    ...
            txt/
                en/
                    dataset.txt
                    ...


If you do not plan to support multiple languages, you can skip the language
directories. In this case, the templates are supposed to be in English language.

To change the contents of a (sub)template for your package, *e.g.* the
"colophon.tex" template containing important information on how a report has
been generated, simply provide a template with this name in the
corresponding template directory within your package. As long as you are
using recipe-driven data analysis and provide a default package in your
recipe, the reporters will be notified and look for templates in your
package before defaulting to the templates provided with the ASpecD framework.


Background: Jinja2
==================

The report functionality relies heavily on the `Jinja2 template engine
<http://jinja.pocoo.org/>`_ . Therefore, it is very useful to be generally
familiar with the concepts of Jinja2. Basically, this template engine
allows users to specify rather complicated replacements and logic within
the template. However, logic within templates should be used sparingly,
and the template should always be rendered correctly even without
processing it by the template engine. Only then templates are easy to
develop and adapt.

Two concepts used by Jinja2 the user of the report facilities of the ASpecD
framework should be familiar with are "environment" and "context". The
former is a list of settings determining the type of delimiters used within
a certain template for the control structures that are understood by
Jinja2. As Jinja2 is developed with web applications (and hence HTML) in
mind, those delimiters may not be feasible for other types of languages a
template may be written in, such as LaTeX.

Currently, the :mod:`aspecd.report` module of the ASpecD framework provides
a generic environment as well as dedicated Txt and LaTeX environments,
implemented as respective classes:

  * :class:`GenericEnvironment`
  * :class:`TxtEnvironment`
  * :class:`LaTeXEnvironment`

These environments get automatically loaded by the respective reporter
classes:

  * :class:`Reporter`
  * :class:`TxtReporter`
  * :class:`LaTeXReporter`

While the :class:`TxtEnvironment` and :class:`TxtReporter` classes are
basically identical to the :class:`GenericEnvironment` and :class:`Reporter`
classes, respectively, the :class:`LaTeXEnvironment` and
:class:`LaTeXReporter` provide some heavy adaptations to rendering and even
compiling LaTeX templates.

The second important concept of Jinja2 is that of the "context": Think of
it as a dictionary containing all the key--value pairs you can use to
replace placeholders within a template with their actual values. In the
simplest of all cases within the context of the ASpecD framework,
this could be the metadata of an :class:`aspecd.dataset.Dataset`.


Module documentation
====================

"""

import collections
import os
import shutil
import subprocess  # nosec
import tempfile
from datetime import datetime

import jinja2
import pkg_resources

import aspecd.exceptions
import aspecd.system
import aspecd.utils


class Reporter(aspecd.utils.ToDictMixin):
    """Base class for reports.

    To generate a report from a template, you will need a template in the
    first place. The path to the template (including its filename) is set in
    the :attr:`template` attribute that is either provided as parameter at
    initialisation or assigned later on. Similarly, you should specify an
    output filename, stored in the :attr:`filename` attribute. The variables
    known to the template that should be replaced by appropriate content,
    termed "context" in Jinja2 language, are stored in the :attr:`context`
    attribute. This is, by default, an ordered dict to preserve the order of
    the keys it contains. This is sometimes very useful.

    Next, you want to render the template, and most often, save it to a file
    afterwards. If you would like to separate those steps, you can call the
    :meth:`render` and :meth:`save` methods separately. For convenience,
    simply call :meth:`create` to render the report and save it to the file
    whose name you have provided.

    The whole procedure may look as follows::

        template = "/path/to/my/template.tex"
        filename = "/path/to/my/final/report.tex"
        report_ = aspecd.report.Reporter(template=template, filename=filename)
        report_.create()


    Attributes
    ----------
    template : :class:`str`
        Template file used to generate report.

    filename : :class:`str`
        Name of the resulting template file.

    context : :class:`collections.OrderedDict`
        Variables of a template that are replaced with the given content.

        It contains a key "sysinfo" containing system-related
        information, i.e. the information contained in the
        :class:`aspecd.system.SystemInfo` class.

        Furthermore, the key "template_dir" contains the (relative or
        absolute) path to the template provided in :attr:`template`. This is
        particularly useful for including subtemplates.

        A key "timestamp" contains the current timestamp when starting to
        render the report.

    environment : :class:`aspecd.report.GenericEnvironment`
        Jinja2 environment used for rendering the template.

        Defaults to a :obj:`aspecd.report.GenericEnvironment` object with
        settings for rendering generic templates.

    report : :class:`str`
        Actual report, i.e. rendered template

    package : :class:`str`
        Name of the package a template loader shall be added for

        Additionally to adding a template loader, the package name and its
        version number as well as all its dependencies get added to the
        ``sysinfo`` key of the :attr:`context` dictionary.

    package_path : :class:`str`
        Path to the templates within the package defined by :attr:`package`

    language : :class:`str`
        (Human) language of the templates

        Usually a two-letter code. Only if the corresponding template
        directory exists within the package, the language will be set.

    comment : :class:`str`
        User-supplied comment describing intent, purpose, reason, ...


    Parameters
    ----------
    template : :class:`str`
        Path to template file used to generate report.

    filename : :class:`str`
        Path to the output file the report should be rendered to.

    Raises
    ------
    aspecd.report.FileNotFoundError
        Raised if the template file provided does not exist.

    aspecd.report.MissingFilenameError
        Raised if no output file for the report is provided.


    .. versionchanged:: 0.6.3
        New attributes :attr:`package`, :attr:`package_path`, :attr:`language`

    .. versionchanged:: 0.6.4
        New attribute :attr:`comment`

    """

    def __init__(self, template="", filename=""):
        super().__init__()
        self.template = template
        self.filename = filename
        self.context = collections.OrderedDict()
        self.environment = GenericEnvironment()
        self.report = ""
        self.package = ""
        self.package_path = ""
        self.language = ""
        self.comment = ""
        self._jinja_template = None
        self.__kind__ = "report"
        self._exclude_from_to_dict = [
            "context",
            "environment",
            "report",
            "package",
            "package_path",
        ]

    def render(self):
        """Render the template.

        The actual rendering of the template should be implemented within the
        non-public method :meth:`_render`.

        Before calling the non-public method :meth:`_render`,
        the renderer checks the existence of the file provided as attribute
        :attr:`template` as well as whether an output filename has been
        provided. If this is not the case, a corresponding exception will be
        raised.

        Raises
        ------
        FileNotFoundError
            Raised if the template file provided does not exist.

        """
        if not self.template:
            raise FileNotFoundError("No template provided")
        # noinspection PyTypeChecker
        self._add_package_loader()
        self._set_language()
        self._add_to_context()
        self._get_jinja_template()
        self._render()

    def _add_package_loader(self):
        if self.package:
            if self.package_path:
                self.environment.add_package_loader(
                    package_name=self.package, package_path=self.package_path
                )
            else:
                self.environment.add_package_loader(package_name=self.package)

    def _set_language(self):
        if self.language:
            self.environment.language = self.language
            self.environment.set_language()

    def _add_to_context(self):
        self.context["sysinfo"] = aspecd.system.SystemInfo(
            package=self.package
        ).to_dict()
        self.context["template_dir"] = os.path.split(self.template)[0]
        if self.context["template_dir"]:
            self.context["template_dir"] += os.path.sep
        # noinspection PyTypeChecker
        self.context["timestamp"] = datetime.now().strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    def _get_jinja_template(self):
        try:
            self._jinja_template = self.environment.get_template(
                self.template
            )
        except jinja2.exceptions.TemplateError:
            self.template = os.path.realpath(self.template)
            self._jinja_template = self.environment.get_template(
                self.template
            )

    def _render(self):
        """Perform the actual rendering of the template.

        The implementation of the actual rendering goes in here in all
        classes inheriting from Report. This method is automatically
        called by :meth:`render` after some background checks.

        By default, the Jinja2 environment set in the :attr:`environment`
        attribute is used together with the template provided in the
        :attr:`template` attribute and the context provided in the
        :attr:`context` attribute. The result of the rendering is stored in
        the :attr:`report` attribute.

        If you need to change anything here, simply override this method in
        a child class according to your needs.
        """
        self.report = self._jinja_template.render(self.context)

    def save(self):
        """Save report to file.

        Raises
        ------
        aspecd.report.MissingFilenameError
            Raised if no output file for the report is provided.

        """
        if not self.filename:
            raise aspecd.exceptions.MissingFilenameError(
                "No output file for report"
            )
        with open(self.filename, mode="w+", encoding="utf8") as output_file:
            output_file.write(self.report)

    def create(self):
        """Create report, thus rendering template and saving result to file.

        Convenience method to render the template and save the generated
        report to a file, thus simply calling :meth:`render` and
        :meth:`save` subsequently.

        """
        self.render()
        self.save()


class TxtReporter(Reporter):
    """Plain text reporter.

    The most basic format for a report is plain text. Its probably biggest
    advantage is that it is intrinsically platform-independent (except of
    the encoding, but UTF-8 should not pose any problems any more in 2021
    and after).

    The perhaps biggest disadvantage is the lack of standard formatting and
    the overall limited formatting options, not to speak of including figures.

    As such, the plain text reporter can be used for a first overview and
    for maximum portability. For well-formatted reports, have a look at the
    :class:`LaTeXReporter`.

    Attributes
    ----------
    environment : :class:`aspecd.report.TxtEnvironment`
        Jinja2 environment used for rendering the template.

        Similar to the :class:`aspecd.report.GenericEnvironment`, but with
        the package path for template lookup set to the path for txt
        templates within the ASpecD package.


    .. versionadded:: 0.6

    """

    def __init__(self, template="", filename=""):
        super().__init__(template=template, filename=filename)
        self.environment = TxtEnvironment()


class LaTeXReporter(Reporter):
    """LaTeX Reporter.

    Often, templates for reports are written in LaTeX, and the results
    typeset as PDF file upon a (pdf)LaTeX run. For convenience, this class
    offers the necessary facilities to compile the template once written.

    The whole procedure may look as follows::

        template = "template.tex"
        filename = "report.tex"
        report_ = aspecd.report.LaTeXReporter(template=template,
                                              filename=filename)
        report_.create()
        report_.compile()

    This will result with a file "report.pdf" in the current directory. If
    you specify a relative or absolute path for the filename of the report,
    the resulting PDF file will be copied to that path accordingly.

    Note that for compiling a temporary directory is used, such as not to
    clutter the current working directory with all the auxiliary files
    usually created during a (pdf)LaTeX run. Furthermore, currently, only a
    single (pdf)LaTeX run is performed with option
    "-interaction=nonstopmode" passed in order to not block further execution.

    .. important::
        For enhanced security, the temporary directory used for compiling
        the template will be removed after successful compilation.
        Therefore, no traces of your report should remain outside the
        current directory controlled by the user.

    .. note::
        Due to problems with LaTeX rendering text containing underscores,
        the keys in the context dict are recursively parsed and each
        key containing underscores converted to camel case (but preserving
        the case of the first character: "foo_bar" => "fooBar"). Thus,
        the template can be compiled using LaTeX without having to replace
        the placeholder variables beforehand.

    Attributes
    ----------
    environment : :class:`aspecd.report.LaTeXEnvironment`
        Jinja2 environment used for rendering the template.

        Defaults to a :obj:`aspecd.report.LaTeXEnvironment` object with
        settings for rendering LaTeX templates.

    includes : :class:`list`
        List of files that need to be present for compiling the template.

        These files will be copied into the temporary directory used for
        compiling the template.

    latex_executable : :class:`str`
        Name of/path to the LaTeX executable.

        Defaults to "pdflatex"

    Parameters
    ----------
    template : :class:`str`
        Path to template file used to generate report.

    filename : :class:`str`
        Path to the output file the report should be rendered to.

    Raises
    ------
    aspecd.report.LaTeXExecutableNotFoundError
        Raised if the LaTeX executable could not be found

    """

    def __init__(self, template="", filename=""):
        super().__init__(template=template, filename=filename)
        self.environment = LaTeXEnvironment()
        self.includes = []
        self.latex_executable = "pdflatex"

        self._temp_dir = tempfile.mkdtemp()
        self._pwd = os.getcwd()
        self._exclude_from_to_dict.extend(["latex_executable"])

    def _render(self):
        """Perform the actual rendering of the template.

        Additionally to the usual tasks performed in the base class,
        for LaTeX templates, each key containing underscores is converted to
        camel case (but preserving the case of the first character:
        "foo_bar" => "fooBar"). Thus, the template can be compiled using
        LaTeX without having to replace the placeholder variables beforehand.

        """
        self.context = self._change_keys_in_dict_recursively(self.context)
        super()._render()

    def _change_keys_in_dict_recursively(self, dict_=None):
        tmp_dict = collections.OrderedDict()
        for key, value in dict_.items():
            if isinstance(value, dict):
                dict_[key] = self._change_keys_in_dict_recursively(value)
            if "_" in key:
                tmp_key = "".join(
                    [x.capitalize() for x in key.split(sep="_")]
                )
                tmp_key = "".join([tmp_key[0].lower(), tmp_key[1:]])
            else:
                tmp_key = key
            tmp_dict[tmp_key] = dict_[key]
        return tmp_dict

    def compile(self):
        """Compile LaTeX template.

        The template is copied to a temporary directory and the LaTeX
        executable specified in :attr:`latex_executable` called on the
        report. Afterwards, the result is copied back to the original
        directory.

        Additionally, all files necessary to compile the report are copied
        to the temporary directory as well.

        Raises
        ------
        aspecd.report.LaTeXExecutableNotFoundError
            Raised if the LaTeX executable could not be found

        """
        if not shutil.which(self.latex_executable):
            raise aspecd.exceptions.LaTeXExecutableNotFoundError
        self._copy_files_to_temp_dir()
        self._compile()
        # Note: In order to resolve references in LaTeX, compile twice.
        #       There might be a better option, automatically detecting
        #       whether compiling twice is necessary, but for now...
        self._compile()
        self._copy_files_from_temp_dir()
        self._remove_temp_dir()

    def _copy_files_to_temp_dir(self):
        """Copy all necessary files to compile the LaTeX report to temp_dir

        Takes care of relative or absolute paths of both, report and includes.
        """
        _, filename_wo_path = os.path.split(self.filename)
        shutil.copy2(
            self.filename, os.path.join(self._temp_dir, filename_wo_path)
        )
        for filename in self.includes:
            _, filename_wo_path = os.path.split(filename)
            shutil.copy2(
                filename, os.path.join(self._temp_dir, filename_wo_path)
            )

    def _compile(self):
        """Actual compiling of the report.

        The compiling takes place in a temporary directory that gets
        removed after the (successful) compile step using the
        :meth:`_remove_temp_dir` method.

        (pdf)LaTeX is currently called with the "-interaction=nonstopmode"
        option in order to not block further execution.
        """
        with aspecd.utils.change_working_dir(self._temp_dir):
            _, filename_wo_path = os.path.split(self.filename)
            # Path stripped, there should be no security implications.
            process = subprocess.run(
                [
                    self.latex_executable,  # nosec
                    "-output-directory",
                    self._temp_dir,
                    "-interaction=nonstopmode",
                    filename_wo_path,
                ],
                check=False,
                capture_output=True,
            )
            print(process.stdout.decode())
            print(process.stderr.decode())

    def _copy_files_from_temp_dir(self):
        """Copy result of compile step from temporary to target directory

        Takes care of any relative or absolute paths provided for the
        report output file provided in :attr:`filename`.
        """
        basename, _ = os.path.splitext(self.filename)
        path, basename = os.path.split(basename)
        pdf_filename = ".".join([basename, "pdf"])
        if os.path.isabs(path):
            shutil.copy2(
                os.path.join(self._temp_dir, pdf_filename),
                os.path.join(path, pdf_filename),
            )
        else:
            shutil.copy2(
                os.path.join(self._temp_dir, pdf_filename),
                os.path.join(self._pwd, path, pdf_filename),
            )

    def _remove_temp_dir(self):
        """Remove temporary directory used for compile step.

        The (pdf)LaTeX step is performed in a temporary directory such as
        not to clutter the current working directory with all the auxiliary
        files usually created during a (pdf)LaTeX run.

        This temporary directory is removed after the (successful) compile
        step. Note that therefore, all sometimes useful information stored,
        e.g., in the log file, is lost. However, manually compiling the
        report is probably the easiest way of figuring out if something
        gets wrong with the (pdf)LaTeX compile step
        """
        shutil.rmtree(self._temp_dir)


class GenericEnvironment(jinja2.Environment):
    """Jinja2 environment for rendering generic templates.

    The environment does not change any of the jinja settings except of the
    loaders. Here, a list of loaders using :class:`jinja2.ChoiceLoader` is
    implemented. Using this loader makes it possible to search subsequently
    in different places for a template. Here, the first hit is used,
    therefore, the sequence of loaders is *crucial*.

    Currently, there are two loaders implemented, in exactly this sequence:

    #. :class:`jinja2.FileSystemLoader`

        Looking for templates in the current directory and using an absolute
        path

    #. :class:`jinja2.PackageLoader`

        Looking for templates in the aspecd package in the package path
        "templates/report/", *i.e.* the base directory for all report
        templates of the ASpecD package.

    Additional package loaders can be inserted before the package loader for
    the ASpecD package using the method :meth:`add_package_loader`. This
    allows packages based on the ASpecD framework to define templates with
    the same name as those in ASpecD and thus to load the templates provided
    by the derived package rather than those from ASpecD.


    Attributes
    ----------
    path : :class:`str`
        Path to the templates within the package

        Default: "templates/report"


    Parameters
    ----------
    env : :class:`dict`
        Dictionary used for creating the :class:`jinja2.Environment`

        Can be used by derived classes to override the environment.

    path : :class:`str`
        Path to the templates within the package

    lang : :class:`str`
        Language of the templates

        Usually a two-letter code; if a corresponding subdirectory to
        :attr:`path` exists within the template package directory, it will be
        added to the path of the template package loaders.


    .. versionchanged:: 0.6.3
        New attribute :attr:`package_path`, new parameters ``env``,
        ``path``, ``lang``

    """

    def __init__(self, env=None, path="templates/report/", lang=None):
        self.path = path
        self.language = lang
        if not env:
            env = {
                "loader": jinja2.ChoiceLoader(
                    [
                        jinja2.FileSystemLoader(
                            [os.path.abspath("."), os.path.abspath("/")]
                        ),
                        jinja2.PackageLoader(
                            "aspecd", package_path=self.path
                        ),
                    ]
                )
            }
        super().__init__(**env)
        self.set_language()

    def set_language(self):
        """
        Adjust the template directory of the package loaders for the language.

        Only if a language is set in the class and the corresponding
        template directory exists, the loaders will be updated. As it seems,
        there is currently no way to adjust the package_path of a
        :class:`jinja2.PackageLoader`, hence the loaders are replaced with
        new loaders with adjusted package_path.

        """
        if self.language:
            for idx, loader in enumerate(self.loader.loaders):
                if isinstance(loader, jinja2.PackageLoader):
                    package_name = loader.package_name
                    package_path = loader.package_path
                    if package_path.endswith("en"):
                        package_path = package_path.rstrip("/en")
                    package_path = "/".join([package_path, self.language])
                    if pkg_resources.resource_exists(
                        package_name, package_path
                    ):
                        self.loader.loaders[idx] = jinja2.PackageLoader(
                            package_name=package_name,
                            package_path=package_path,
                        )

    def add_package_loader(self, package_name="", package_path=""):
        """
        Add a package loader for a given package name.

        The package loader will be inserted before the loader for the ASpecD
        package. This allows packages based on the ASpecD framework to
        define templates with the same name as those in ASpecD and thus to
        load the templates provided by the derived package rather than those
        from ASpecD.

        Only in case of the given package path to exist the package loader
        will be added.


        Parameters
        ----------
        package_name : :class:`str`
            Name of the package to add the loader for

        package_path : :class:`str`
            Path to the templates within the package

            Defaults to :attr:`package_path`


        .. versionadded:: 0.6.3

        """
        if not package_path:
            package_path = self.path
        if self.language:
            package_path = "/".join([package_path, self.language])
        if pkg_resources.resource_exists(package_name, package_path):
            package_loader = jinja2.PackageLoader(
                package_name=package_name, package_path=package_path
            )
            self.loader.loaders.insert(-1, package_loader)


class TxtEnvironment(GenericEnvironment):
    """Jinja2 environment for rendering generic text templates.

    The environment does not change any of the jinja settings except of the
    loaders. Here, a list of loaders using :class:`jinja2.ChoiceLoader` is
    implemented. Using this loader makes it possible to search subsequently
    in different places for a template. Here, the first hit is used,
    therefore, the sequence of loaders is *crucial*.

    Currently, there are two loaders implemented, in exactly this sequence:

    #. :class:`jinja2.FileSystemLoader`

        Looking for templates in the current directory and using an absolute
        path

    #. :class:`jinja2.PackageLoader`

        Looking for templates in the aspecd package in the package path
        "templates/report/txt/", *i.e.* the directory for all bare text report
        templates of the ASpecD package.


    .. versionchanged:: 0.6.3
        Now based on :class:`GenericEnvironment`

    """

    def __init__(self, lang=None):
        self.package_path = "templates/report/txt/"
        self.lang = lang or "en"
        env = {
            "loader": jinja2.ChoiceLoader(
                [
                    jinja2.FileSystemLoader(
                        [os.path.abspath("."), os.path.abspath("/")]
                    ),
                    jinja2.PackageLoader(
                        "aspecd", package_path=self.package_path
                    ),
                ]
            )
        }
        super().__init__(env=env, path=self.package_path, lang=self.lang)


class LaTeXEnvironment(GenericEnvironment):
    """Jinja2 environment for rendering LaTeX-based templates.

    This environment is designed for using templates written in LaTeX that
    can be rendered by LaTeX without having their control code replaced by
    Jinja2. While variables are usually output in LaTeX, control structures
    are prefixed by a LaTeX comment character (``%``). For convenience,
    the following table lists all the variables currently set within this
    environment.

    ===================== =====
    Variable              Value
    ===================== =====
    block_start_string    %{
    block_end_string      }%
    variable_start_string {@
    variable_end_string   }
    comment_start_string  %#{
    comment_end_string    }
    line_statement_prefix %%
    line_comment_prefix   %#
    trim_blocks           True
    autoescape            False
    ===================== =====

    While every measure is taken to keep the above information as accurate
    as possible, for authoritative information the reader is referred to
    the actual source code.

    Besides extensively modifying the control codes used within the
    template, the environment implements a list of loaders using
    :class:`jinja2.ChoiceLoader`. Using this loader makes it possible to
    search subsequently in different places for a template. Here, the first
    hit is used, therefore, the sequence of loaders is *crucial*.

    Currently, there are two loaders implemented, in exactly this sequence:

    #. :class:`jinja2.FileSystemLoader`

        Looking for templates in the current directory and using an absolute
        path

    #. :class:`jinja2.PackageLoader`

        Looking for templates in the aspecd package in the package path
        "templates/report/latex/", *i.e.* the directory for all bare text report
        templates of the ASpecD package.


    .. versionchanged:: 0.6.3
        Now based on :class:`GenericEnvironment`

    """

    def __init__(self, lang=None):
        self.package_path = "templates/report/latex/"
        self.lang = lang or "en"
        env = {
            "block_start_string": "%{",
            "block_end_string": "}%",
            "variable_start_string": "{@",
            "variable_end_string": "}",
            "comment_start_string": "%#{",
            "comment_end_string": "}",
            "line_statement_prefix": "%%",
            "line_comment_prefix": "%#",
            "trim_blocks": True,
            "autoescape": False,
            "loader": jinja2.ChoiceLoader(
                [
                    jinja2.FileSystemLoader(
                        [os.path.abspath("."), os.path.abspath("/")]
                    ),
                    jinja2.PackageLoader(
                        "aspecd", package_path=self.package_path
                    ),
                ]
            ),
        }
        super().__init__(env=env, path=self.package_path, lang=self.lang)



"""
Handle Infofile files containing metadata for (experimental) data.

The Infofile format is a special format dedicated to storing metadata
accompanying experimental data. Further information on the format
including the full specification can be found in the following publication:

  * Bernd Paulus, Till Biskup: Towards more reproducible and FAIRer
    research data: documenting provenance during data acquisition using the
    Infofile format.
    *Digital Discovery* **2**:234–244, 2023,
    doi:`10.1039/D2DD00131D <https://doi.org/10.1039/D2DD00131D>`_

For actual Infofile example files, see the git repository available at GitHub:

  * https://github.com/tillbiskup/infofile

Development of the templates for specific methods will always be closely
connected to the respective data model, *e.g.* in context of the `trepr
<https://docs.trepr.de/>`_ and `cwepr <https://docs.cwepr.de/>`_ Python
packages.

This module reads and parses files complying with the infofile file
format. At the same time, it contains the **reference implementation of
the parser for the Infofile format**.

.. note::
    For own purposes, you may prefer using YAML files to store metadata
    that are collected in parallel to recording data over the infofile
    format. For more information on this format and it specification,
    consult its webpage: `<https://yaml.org/>`_.

    However, you may still be interested in the discussion of the Infofile
    format and its advantages and disadvantages compared to formats such
    as YAML and JSON. For this, you are kindly referred to the reference
    given above, describing and discussing the Infofile format in more detail.

    As the lead developer of the ASpecD framework originally invented the
    infofile format for own purposes, there are quite some datasets floating
    around in his lab using this format for the corresponding metadata.
    Hence the need of a module handling this particular file type.

To map the information contained in the info file to datasets, have a look
at the :class:`aspecd.metadata.MetadataMapper` class. You can define rather
elaborate mapping tables, allowing to rename keys as well as to combine
items of a dictionary.
"""

import collections
import os
import string

import aspecd.exceptions


class Infofile:
    """Reading metadata contained in info files.

    Attributes
    ----------
    parameters : :class:`collections.OrderedDict`
        Structure containing parameters read from info file.
    filename : :class:`str`
        Name of the info file read
    infofile_info : :class:`dict`
        Information about the infofile, such as kind, version, and date.

        Helpful for mapping metadata contained in an infofile to datasets.

    Parameters
    ----------
    filename : :class:`str`
        Name of the info file to read

    Raises
    ------
    aspecd.exceptions.InfofileEmptyError
        Raised if info file is empty
    aspecd.exceptions.InfofileTypeError
        Raised if file provided is no info file


    .. versionchanged:: 0.6.3
        Comment gets converted into a single string.

    """

    def __init__(self, filename=None):
        self.parameters = collections.OrderedDict()
        self.filename = filename
        self.infofile_info = {}

        self.infofile_info["kind"] = ""
        self.infofile_info["version"] = ""
        self.infofile_info["date"] = ""

        self._file_contents = []
        self._comment_char = "%"
        self._escape_char = "\\"

    def parse(self):
        """Parse info file.

        Raises
        ------
        aspecd.exceptions.InfofileEmptyError
            Raised if info file is empty
        aspecd.exceptions.InfofileTypeError
            Raised if file provided is no info file

        """
        self._file_contents = self._read()

        if not self._file_contents:
            raise aspecd.exceptions.InfofileEmptyError

        if not self._is_infofile():
            raise aspecd.exceptions.InfofileTypeError

        self._parse_infofile_info()
        self._parse_infofile_body()

    def _parse_infofile_body(self):  # noqa: MC0001
        blockname = ""
        key = ""
        value = ""
        tmp_block = collections.OrderedDict()

        # Start with second line of file, omitting format and version info
        for line in self._file_contents[1:]:
            if self._is_comment_block(blockname):
                if blockname not in self.parameters:
                    self.parameters[blockname] = line.rstrip()
                else:
                    self.parameters[blockname] = " ".join(
                        [self.parameters[blockname], line.rstrip()]
                    )
                continue
            if self._is_comment_line(line):
                continue
            line = self._remove_inline_comment(line)
            if self._is_continuation_line(line):
                if key:
                    tmp_block[key] = " ".join([value, line.strip()])
                continue
            line = line.strip()
            if not line:
                if blockname:
                    self.parameters[blockname] = tmp_block
                    blockname = ""
                    tmp_block = collections.OrderedDict()
                continue
            if not blockname:
                blockname = line.strip()
                continue
            [key, value] = [i.strip() for i in line.split(":", maxsplit=1)]
            tmp_block[key] = value
        # In case last block has not been assigned
        # (file didn't end with COMMENT block or empty line)
        if blockname:
            if not self._is_comment_block(blockname):
                self.parameters[blockname] = tmp_block
            # noinspection PyUnboundLocalVariable
            # pylint: disable=undefined-loop-variable
            if self._is_comment_block(blockname) and self._is_comment_block(
                line
            ):
                self.parameters[blockname] = ""

    def _is_infofile(self):
        return "info file" in self._file_contents[0].lower()

    def _read(self):
        if not self.filename:
            raise FileExistsError
        if not os.path.exists(self.filename):
            raise FileNotFoundError
        with open(self.filename, encoding="utf8") as file:
            file_contents = list(file)
        return file_contents

    def _parse_infofile_info(self):
        info_line = self._file_contents[0]
        info_parts = info_line.split(" - ")
        _, version, version_date = info_parts[1].split()
        self.infofile_info["kind"] = info_parts[0].strip()
        self.infofile_info["version"] = version
        self.infofile_info["date"] = version_date[1:-1]

    def _is_comment_line(self, line):
        return line.strip().startswith(self._comment_char)

    def _remove_inline_comment(self, line):
        # Get all occurrences (positions) of COMMENTCHAR in line
        occur = [i for i, j in enumerate(line) if j == self._comment_char]
        # Remove those occurrences prepended with ESCAPE character
        occur = [i for i in occur if line[i - 1] != self._escape_char]
        # If there is an inline comment, remove it from line
        if occur:
            line = line[: occur[0]]
        # Replace escaped comment char with comment char
        return line.replace(
            "".join([self._escape_char, self._comment_char]),
            self._comment_char,
        )

    @staticmethod
    def _is_continuation_line(line):
        return line[0] in string.whitespace and line.strip()

    @staticmethod
    def _is_comment_block(blockname):
        return blockname.lower() == "comment"


def parse(filename=""):
    """Parse info file.

    Conventional interface provided for convenience that gives easy access
    to the metadata stored in an infofile. For full use of the capabilities
    of the :class:`Infofile` class, use its object-oriented interface.

    .. warning::
        This will only return the metadata read from the info file, not the
        version information that is read and parsed as well. To obtain this
        version information, e.g., for mapping of metadata onto other
        structures, you should use the object-oriented interface of the
        :class:`Infofile` class.

    Parameters
    ----------
    filename : :class:`str`
        Name of the info file to parse

    Returns
    -------
    metadata : :class:`dict`
        Dictionary with metadata read from info file

    """
    ifile = Infofile(filename)
    ifile.parse()
    return ifile.parameters



"""Tests for input and output (IO)."""

import copy
import os
import unittest
import zipfile

import numpy as np

import aspecd.exceptions
import aspecd.processing
from aspecd import io, dataset, tasks, utils, history


class TestDatasetImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.DatasetImporter()

    def test_instantiate_class(self):
        pass

    def test_instantiate_with_source_sets_source(self):
        source = "filename"
        importer = io.DatasetImporter(source)
        self.assertEqual(importer.source, source)

    def test_has_import_into_method(self):
        self.assertTrue(hasattr(self.importer, "import_into"))
        self.assertTrue(callable(self.importer.import_into))

    def test_import_into_without_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.importer.import_into()

    def test_import_into_with_dataset_sets_dataset(self):
        test_dataset = dataset.Dataset()
        self.importer.source = "filename"
        self.importer.import_into(test_dataset)
        self.assertIs(self.importer.dataset, test_dataset)

    def test_import_into_with_dataset_sets_id(self):
        test_dataset = dataset.Dataset()
        self.importer.source = "filename"
        self.importer.import_into(test_dataset)
        self.assertEqual(test_dataset.id, self.importer.source)

    def test_import_into_with_dataset_sets_label(self):
        test_dataset = dataset.Dataset()
        self.importer.source = "filename"
        self.importer.import_into(test_dataset)
        self.assertEqual(test_dataset.label, self.importer.source)

    def test_import_into_sets_label_without_path(self):
        test_dataset = dataset.Dataset()
        self.importer.source = "/path/to/filename.ext"
        self.importer.import_into(test_dataset)
        self.assertEqual(
            test_dataset.label, os.path.split(self.importer.source)[-1]
        )

    def test_import_into_does_not_override_label(self):
        test_dataset = dataset.Dataset()
        test_dataset.label = "foobar"
        self.importer.source = "/path/to/filename.ext"
        self.importer.import_into(test_dataset)
        self.assertEqual("foobar", test_dataset.label)


class TestDatasetExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.DatasetExporter()

    def test_instantiate_class(self):
        pass

    def test_has_comment_property(self):
        self.assertTrue(hasattr(self.exporter, "comment"))

    def test_instantiate_with_target_sets_target(self):
        target = "filename"
        exporter = io.DatasetExporter(target=target)
        self.assertEqual(exporter.target, target)

    def test_has_export_from_method(self):
        self.assertTrue(hasattr(self.exporter, "export_from"))
        self.assertTrue(callable(self.exporter.export_from))

    def test_export_from_without_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.exporter.export_from()

    def test_export_from_with_dataset_sets_dataset(self):
        test_dataset = dataset.Dataset()
        self.exporter.export_from(test_dataset)
        self.assertIs(self.exporter.dataset, test_dataset)

    def test_has_create_history_record_method(self):
        self.assertTrue(hasattr(self.exporter, "create_history_record"))
        self.assertTrue(callable(self.exporter.create_history_record))

    def test_create_history_record_returns_history_record(self):
        self.exporter.dataset = aspecd.dataset.Dataset()
        history_record = self.exporter.create_history_record()
        self.assertTrue(
            isinstance(
                history_record, aspecd.history.DatasetExporterHistoryRecord
            )
        )


class TestDatasetImporterFactory(unittest.TestCase):
    def setUp(self):
        self.factory = io.DatasetImporterFactory()
        self.source = "/foo"

    def test_instantiate_class(self):
        pass

    def test_get_importer_returns_importer(self):
        importer = self.factory.get_importer(source=self.source)
        self.assertTrue(isinstance(importer, io.DatasetImporter))

    def test_get_importer_sets_source_in_importer(self):
        importer = self.factory.get_importer(source=self.source)
        self.assertEqual(self.source, importer.source)

    def test_get_importer_with_relative_source_sets_absolute_path(self):
        source = "foo"
        root_path = os.path.abspath(os.curdir)
        importer = self.factory.get_importer(source=source)
        self.assertEqual(os.path.join(root_path, source), importer.source)

    def test_get_importer_without_source_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingSourceError):
            self.factory.get_importer()

    def test_get_importer_with_adf_extension_returns_adf_importer(self):
        source = "/foo.adf"
        importer = self.factory.get_importer(source=source)
        self.assertTrue(isinstance(importer, io.AdfImporter))

    def test_get_importer_with_asdf_extension_returns_asdf_importer(self):
        source = "/foo.asdf"
        importer = self.factory.get_importer(source=source)
        self.assertTrue(isinstance(importer, io.AsdfImporter))

    def test_get_importer_with_txt_extension_returns_txt_importer(self):
        source = "/foo.txt"
        importer = self.factory.get_importer(source=source)
        self.assertTrue(isinstance(importer, io.TxtImporter))

    def test_get_importer_with_importer_returns_specific_importer(self):
        source = "/foo"
        importer = "TxtImporter"
        importer = self.factory.get_importer(source=source, importer=importer)
        self.assertTrue(isinstance(importer, io.TxtImporter))

    def test_get_importer_with_parameters_sets_parameters(self):
        source = "/foo"
        parameters = {"foo": "bla", "bar": "blub"}
        importer = self.factory.get_importer(
            source=source, parameters=parameters
        )
        self.assertDictEqual(parameters, importer.parameters)

    def test_returning_abstract_importer_logs_warning(self):
        with self.assertLogs(__package__, level="WARNING") as captured:
            self.factory.get_importer(source=self.source)
        self.assertEqual(len(captured.records), 1)
        self.assertIn("default importer", captured.output[0].lower())


class TestRecipeImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.RecipeImporter()
        self.source = "filename"

    def test_instantiate_class(self):
        pass

    def test_instantiate_with_source_sets_source(self):
        importer = io.RecipeImporter(source=self.source)
        self.assertEqual(importer.source, self.source)

    def test_has_import_into_method(self):
        self.assertTrue(hasattr(self.importer, "import_into"))
        self.assertTrue(callable(self.importer.import_into))

    def test_import_into_without_recipe_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingRecipeError):
            self.importer.import_into()

    def test_import_into_without_parameter_but_recipe_preset(self):
        recipe = tasks.Recipe()
        self.importer.recipe = recipe
        self.importer.import_into()
        self.assertIs(self.importer.recipe, recipe)

    def test_import_into_with_recipe_sets_recipe(self):
        recipe = tasks.Recipe()
        self.importer.import_into(recipe=recipe)
        self.assertIs(self.importer.recipe, recipe)

    def test_import_into_with_source_sets_filename_in_recipe(self):
        recipe = tasks.Recipe()
        importer = io.RecipeImporter(source=self.source)
        importer.recipe = recipe
        importer.import_into()
        self.assertEqual(importer.recipe.filename, self.source)


class TestRecipeExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.RecipeExporter()
        self.target = "filename"

    def test_instantiate_class(self):
        pass

    def test_instantiate_with_target_sets_target(self):
        exporter = io.RecipeExporter(target=self.target)
        self.assertEqual(exporter.target, self.target)

    def test_has_export_from_method(self):
        self.assertTrue(hasattr(self.exporter, "export_from"))
        self.assertTrue(callable(self.exporter.export_from))

    def test_export_from_without_recipe_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingRecipeError):
            self.exporter.export_from()

    def test_export_from_with_recipe_sets_recipe(self):
        recipe = tasks.Recipe()
        self.exporter.export_from(recipe)
        self.assertIs(self.exporter.recipe, recipe)


class TestRecipeYamlImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.RecipeYamlImporter()
        self.recipe_filename = "filename"
        self.recipe_dict = dict()

    def tearDown(self):
        if os.path.exists(self.recipe_filename):
            os.remove(self.recipe_filename)

    def create_recipe(self):
        self.recipe_dict = {
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "singleplot",
                    "type": "SinglePlotter",
                    "properties": {"filename": "foo"},
                }
            ],
        }
        yaml = utils.Yaml()
        yaml.dict = self.recipe_dict
        yaml.write_to(self.recipe_filename)

    def create_recipe_with_numpy_array(self):
        self.recipe_dict = {
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "processing",
                    "type": "ScalarAlgebra",
                    "properties": {"value": np.random.random(1)},
                }
            ],
        }
        yaml = utils.Yaml()
        yaml.dict = copy.deepcopy(self.recipe_dict)
        yaml.serialise_numpy_arrays()
        yaml.write_to(self.recipe_filename)

    def create_recipe_from_dict(self):
        yaml = utils.Yaml()
        yaml.dict = self.recipe_dict
        yaml.write_to(self.recipe_filename)

    def test_instantiate_class(self):
        pass

    def test_is_recipe_importer(self):
        self.assertTrue(isinstance(self.importer, io.RecipeImporter))

    def test_import_sets_task_in_recipe(self):
        self.create_recipe()
        self.importer.source = self.recipe_filename
        recipe = tasks.Recipe()
        recipe.import_from(self.importer)
        self.assertTrue(recipe.tasks[0])

    def test_import_with_numpy_array_sets_task_in_recipe(self):
        self.create_recipe_with_numpy_array()
        self.importer.source = self.recipe_filename
        recipe = tasks.Recipe()
        recipe.import_from(self.importer)
        self.assertEqual(
            self.recipe_dict["tasks"][0]["properties"]["value"],
            recipe.tasks[0].properties["value"],
        )

    def test_import_sets_recipe_version(self):
        self.create_recipe()
        self.importer.source = self.recipe_filename
        recipe = tasks.Recipe()
        self.importer.import_into(recipe)
        self.assertTrue(self.importer.recipe_version)

    def test_import_old_version_detects_old_version(self):
        old_keys = [
            "default_package",
            "autosave_plots",
            "output_directory",
            "datasets_source_directory",
        ]
        for key in old_keys:
            with self.subTest(key=key):
                self.recipe_dict = {
                    key: "",
                    "datasets": ["foo"],
                    "tasks": [
                        {
                            "kind": "singleplot",
                            "type": "SinglePlotter",
                            "properties": {"filename": "foo"},
                        }
                    ],
                }
                self.create_recipe_from_dict()
                self.importer.source = self.recipe_filename
                recipe = tasks.Recipe()
                self.importer.import_into(recipe)
                self.assertEqual("0.1", self.importer.recipe_version)

    def test_import_with_explicit_version_sets_recipe_version(self):
        self.recipe_dict = {
            "format": {"version": "0.1"},
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "singleplot",
                    "type": "SinglePlotter",
                    "properties": {"filename": "foo"},
                }
            ],
        }
        self.create_recipe_from_dict()
        self.importer.source = self.recipe_filename
        recipe = tasks.Recipe()
        self.importer.import_into(recipe)
        self.assertEqual(
            self.recipe_dict["format"]["version"],
            self.importer.recipe_version,
        )

    def test_import_version_0_1_converts_to_current_version(self):
        self.recipe_dict = {
            "default_package": "aspecd",
            "autosave_plots": False,
            "output_directory": "/bar",
            "datasets_source_directory": "/foobar",
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "singleplot",
                    "type": "SinglePlotter",
                    "properties": {"filename": "foo"},
                }
            ],
        }
        self.create_recipe_from_dict()
        self.importer.source = self.recipe_filename
        recipe = tasks.Recipe()
        recipe.import_from(self.importer)
        self.assertEqual(
            self.recipe_dict["default_package"],
            recipe.settings["default_package"],
        )
        self.assertEqual(
            self.recipe_dict["autosave_plots"],
            recipe.settings["autosave_plots"],
        )
        self.assertEqual(
            self.recipe_dict["output_directory"], recipe.directories["output"]
        )
        self.assertEqual(
            self.recipe_dict["datasets_source_directory"],
            recipe.directories["datasets_source"],
        )


class TestRecipeYamlExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.RecipeYamlExporter()
        self.recipe_filename = "filename"

    def tearDown(self):
        if os.path.exists(self.recipe_filename):
            os.remove(self.recipe_filename)

    def test_instantiate_class(self):
        pass

    def test_is_recipe_exporter(self):
        self.assertTrue(isinstance(self.exporter, io.RecipeExporter))

    def test_export_can_be_reimported(self):
        recipe_dict = {
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "singleplot",
                    "type": "SinglePlotter",
                    "properties": {"filename": "foo"},
                }
            ],
        }
        recipe = tasks.Recipe()
        recipe.from_dict(recipe_dict)
        self.exporter.target = self.recipe_filename
        recipe.export_to(self.exporter)
        new_recipe = tasks.Recipe()
        importer = io.RecipeYamlImporter(source=self.recipe_filename)
        new_recipe.import_from(importer)
        self.assertTrue(new_recipe.to_dict())

    def test_export_with_numpy_array_can_be_reimported(self):
        recipe_dict = {
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "processing",
                    "type": "ScalarAlgebra",
                    "properties": {"value": np.random.random(1)},
                }
            ],
        }
        recipe = tasks.Recipe()
        recipe.from_dict(recipe_dict)
        self.exporter.target = self.recipe_filename
        recipe.export_to(self.exporter)
        new_recipe = tasks.Recipe()
        importer = io.RecipeYamlImporter(source=self.recipe_filename)
        new_recipe.import_from(importer)
        self.assertTrue(new_recipe.to_dict())

    def test_export_with_small_numpy_array_converts_array_to_list(self):
        recipe_dict = {
            "datasets": ["foo"],
            "tasks": [
                {
                    "kind": "processing",
                    "type": "ScalarAlgebra",
                    "properties": {"value": np.random.random(1)},
                }
            ],
        }
        recipe = tasks.Recipe()
        recipe.from_dict(recipe_dict)
        self.exporter.target = self.recipe_filename
        recipe.export_to(self.exporter)
        yaml = utils.Yaml()
        yaml.read_from(self.recipe_filename)
        self.assertIsInstance(
            yaml.dict["tasks"][0]["properties"]["value"], list
        )


class TestAdfExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.AdfExporter()
        self.dataset = dataset.Dataset()
        self.target = "target"
        self.extension = ".adf"

    def tearDown(self):
        if os.path.exists(self.target + self.extension):
            os.remove(self.target + self.extension)

    def test_instantiate_class(self):
        pass

    def test_export_from_without_target_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingTargetError):
            self.exporter.export_from(dataset=self.dataset)

    def test_export_with_target_creates_file(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        self.assertTrue(os.path.exists(self.target + self.extension))

    def test_export_creates_zip_file(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        self.assertTrue(zipfile.is_zipfile(self.target + self.extension))

    def test_created_zipfile_contains_correct_files(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        expected_list_of_files = [
            "dataset.yaml",
            "VERSION",
            "README",
            "binaryData/",
        ]
        expected_list_of_files.sort()
        with zipfile.ZipFile(
            self.target + self.extension, "r"
        ) as zipped_file:
            actual_list_of_files = zipped_file.namelist()
        actual_list_of_files.sort()
        self.assertListEqual(expected_list_of_files, actual_list_of_files)

    def test_export_with_large_array_adds_binary_files(self):
        self.exporter.target = self.target
        self.dataset.data.data = np.random.random(1001)
        self.exporter.export_from(self.dataset)
        with zipfile.ZipFile(
            self.target + self.extension, "r"
        ) as zipped_file:
            list_of_files = zipped_file.namelist()
        binary_files = [
            x for x in list_of_files if x.startswith("binaryData")
        ]
        self.assertEqual(3, len(binary_files))

    def test_export_sets_version(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        with zipfile.ZipFile(
            self.target + self.extension, "r"
        ) as zipped_file:
            version = zipped_file.read("VERSION")
        self.assertEqual(self.exporter._version, version.decode("ascii"))

    def test_export_writes_readme(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        with zipfile.ZipFile(
            self.target + self.extension, "r"
        ) as zipped_file:
            readme = zipped_file.read("README")
        self.assertTrue(readme)


class TestAdfImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.AdfImporter()
        self.dataset = dataset.ExperimentalDataset()
        self.source = "target"
        self.extension = ".adf"
        self.exporter = io.AdfExporter()
        self.exporter.target = self.source

    def tearDown(self):
        if os.path.exists(self.source + self.extension):
            os.remove(self.source + self.extension)

    def test_instantiate_class(self):
        pass

    def test_import_sets_data_and_origdata(self):
        dataset_ = dataset.Dataset()
        dataset_.data.data = np.asarray([1.0])
        dataset_._origdata.data = np.asarray([1.0])
        dataset_.export_to(self.exporter)

        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertEqual(dataset_.data.data, self.dataset.data.data)
        self.assertEqual(dataset_._origdata.data, self.dataset._origdata.data)

    def test_import_sets_metadata(self):
        dataset_ = dataset.ExperimentalDataset()
        dataset_.metadata.sample.name = "foo"
        dataset_.export_to(self.exporter)
        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertEqual(
            dataset_.metadata.sample.name, self.dataset.metadata.sample.name
        )

    def test_import_sets_history(self):
        dataset_ = dataset.ExperimentalDataset()
        processing_step = aspecd.processing.SingleProcessingStep()
        processing_step.comment = "foo"
        dataset_.process(processing_step)
        dataset_.export_to(self.exporter)
        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertDictEqual(
            dataset_.history[0].to_dict(), self.dataset.history[0].to_dict()
        )

    def test_import_with_large_array(self):
        dataset_ = dataset.Dataset()
        dataset_.data.data = np.random.random(1001)
        dataset_.export_to(self.exporter)

        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        np.testing.assert_allclose(dataset_.data.data, self.dataset.data.data)


class TestAsdfExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.AsdfExporter()
        self.dataset = dataset.Dataset()
        self.target = "target"
        self.extension = ".asdf"

    def tearDown(self):
        if os.path.exists(self.target + self.extension):
            os.remove(self.target + self.extension)

    def test_instantiate_class(self):
        pass

    def test_export_from_without_target_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingTargetError):
            self.exporter.export_from(dataset=self.dataset)

    def test_export_with_target_creates_file(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        self.assertTrue(os.path.exists(self.target + self.extension))


class TestAsdfImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.AsdfImporter()
        self.dataset = dataset.ExperimentalDataset()
        self.source = "target"
        self.extension = ".asdf"
        self.exporter = io.AsdfExporter()
        self.exporter.target = self.source

    def tearDown(self):
        if os.path.exists(self.source + self.extension):
            os.remove(self.source + self.extension)

    def test_instantiate_class(self):
        pass

    def test_import_sets_data_and_origdata(self):
        dataset_ = dataset.Dataset()
        dataset_.data.data = np.asarray([1.0])
        dataset_._origdata.data = np.asarray([1.0])
        dataset_.export_to(self.exporter)

        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertEqual(dataset_.data.data, self.dataset.data.data)
        self.assertEqual(dataset_._origdata.data, self.dataset._origdata.data)

    def test_import_sets_metadata(self):
        dataset_ = dataset.ExperimentalDataset()
        dataset_.metadata.sample.name = "foo"
        dataset_.export_to(self.exporter)
        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertEqual(
            dataset_.metadata.sample.name, self.dataset.metadata.sample.name
        )

    def test_import_sets_history(self):
        dataset_ = dataset.ExperimentalDataset()
        processing_step = aspecd.processing.SingleProcessingStep()
        processing_step.comment = "foo"
        dataset_.process(processing_step)
        dataset_.export_to(self.exporter)
        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertDictEqual(
            dataset_.history[0].to_dict(), self.dataset.history[0].to_dict()
        )


class TestTxtImporter(unittest.TestCase):
    def setUp(self):
        self.importer = io.TxtImporter()
        self.dataset = dataset.ExperimentalDataset()
        self.source = "target.txt"

    def tearDown(self):
        if os.path.exists(self.source):
            os.remove(self.source)

    def test_instantiate_class(self):
        pass

    def test_import_sets_data_and_origdata(self):
        data = np.random.random(5)
        np.savetxt(self.source, data)

        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertListEqual(list(data), list(self.dataset.data.data))
        self.assertListEqual(list(data), list(self.dataset._origdata.data))

    def test_import_two_column_data_sets_data_and_axis(self):
        data = np.random.random([5, 2])
        np.savetxt(self.source, data)

        self.importer.source = self.source
        self.dataset.import_from(self.importer)
        self.assertEqual(1, self.dataset.data.data.ndim)
        np.testing.assert_allclose(data[:, 1], self.dataset.data.data)
        np.testing.assert_allclose(data[:, 1], self.dataset._origdata.data)
        np.testing.assert_allclose(
            data[:, 0], self.dataset.data.axes[0].values
        )
        np.testing.assert_allclose(
            data[:, 0], self.dataset._origdata.axes[0].values
        )

    def test_import_2d_data_sets_data_and_axis(self):
        data = np.random.random([5, 4])
        np.savetxt(self.source, data)

        self.importer.source = self.source
        self.importer.parameters["axis"] = 2
        self.dataset.import_from(self.importer)
        np.testing.assert_allclose(
            np.delete(data, 2, 1), self.dataset.data.data
        )
        np.testing.assert_allclose(
            data[:, 2], self.dataset.data.axes[0].values
        )

    def test_import_with_skiprows(self):
        data = np.random.random([5, 1])
        np.savetxt(self.source, data)

        skiprows = 2
        self.importer.source = self.source
        self.importer.parameters["skiprows"] = skiprows
        self.dataset.import_from(self.importer)
        self.assertEqual(len(data) - skiprows, len(self.dataset.data.data))

    def test_import_with_delimiter(self):
        with open(self.source, "w+") as file:
            file.write("1;5;1\n2;6;2\n3;7;3\n4;8;4\n5;9;5\n")

        self.importer.source = self.source
        self.importer.parameters["delimiter"] = ";"
        self.importer.parameters["axis"] = None
        self.dataset.import_from(self.importer)
        self.assertEqual((5, 3), self.dataset.data.data.shape)

    def test_import_with_comments(self):
        with open(self.source, "w+") as file:
            file.write("% Lorem ipsum\n1 5 1\n2 6 2\n3 7 3\n4 8 4\n5 9 5\n")

        self.importer.source = self.source
        self.importer.parameters["comments"] = "%"
        self.importer.parameters["axis"] = None
        self.dataset.import_from(self.importer)
        self.assertEqual((5, 3), self.dataset.data.data.shape)

    def test_import_with_separator(self):
        with open(self.source, "w+") as file:
            file.write("1,1\n2,2\n3,3\n4,4\n5,5\n")

        self.importer.source = self.source
        self.importer.parameters["separator"] = ","
        self.dataset.import_from(self.importer)
        self.assertEqual((5,), self.dataset.data.data.shape)


class TestTxtExporter(unittest.TestCase):
    def setUp(self):
        self.exporter = io.TxtExporter()
        self.dataset = dataset.ExperimentalDataset()
        self.target = "target.txt"

    def tearDown(self):
        if os.path.exists(self.target):
            os.remove(self.target)

    def test_instantiate_class(self):
        pass

    def test_export_from_without_target_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingTargetError):
            self.exporter.export_from(dataset=self.dataset)

    def test_export_with_target_creates_file(self):
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        self.assertTrue(os.path.exists(self.target))

    def test_export_creates_file_with_correct_extension(self):
        self.exporter.target = os.path.splitext(self.target)[0]
        self.exporter.export_from(self.dataset)
        self.assertTrue(os.path.exists(self.target))

    def test_export_with_1D_dataset_adds_axis_as_first_column(self):
        self.dataset.data.data = np.linspace(3, 4, 11)
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        with open(self.target, "r") as file:
            first_line = file.readline()
        self.assertTrue(first_line.startswith("0."))
        self.assertIn("3.", first_line)

    def test_export_with_3D_dataset_raises(self):
        self.dataset.data.data = np.random.random([3, 3, 3])
        self.exporter.target = self.target
        with self.assertRaises(ValueError):
            self.exporter.export_from(self.dataset)

    def test_export_with_2D_dataset_adds_axes_as_first_column_row(self):
        self.dataset.data.data = np.random.random([3, 4])
        self.exporter.target = self.target
        self.exporter.export_from(self.dataset)
        with open(self.target, "r") as file:
            first_line = file.readline()
            second_line = file.readline()
            third_line = file.readline()
        self.assertTrue(first_line.startswith("0."))
        self.assertIn("3.", first_line)
        self.assertTrue(second_line.startswith("0."))
        self.assertTrue(third_line.startswith("1."))



import copy
import unittest

import numpy as np

import aspecd.dataset
import aspecd.exceptions
import aspecd.utils
from aspecd import model


class TestModel(unittest.TestCase):
    def setUp(self):
        self.model = model.Model()

    def test_instantiate_class(self):
        pass

    def test_has_name_property(self):
        self.assertTrue(hasattr(self.model, "name"))

    def test_name_property_equals_full_class_name(self):
        full_class_name = aspecd.utils.full_class_name(self.model)
        self.assertEqual(self.model.name, full_class_name)

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.model, "parameters"))

    def test_has_variables_property(self):
        self.assertTrue(hasattr(self.model, "variables"))

    def test_has_description_property(self):
        self.assertTrue(hasattr(self.model, "description"))

    def test_description_property_describes_abstract_model(self):
        self.assertIn("abstract model", self.model.description.lower())

    def test_has_references_property(self):
        self.assertTrue(hasattr(self.model, "references"))

    def test_description_references_is_list(self):
        self.assertTrue(isinstance(self.model.references, list))

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.model, "to_dict"))
        self.assertTrue(callable(self.model.to_dict))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["name", "description", "references"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.model.to_dict())

    def test_has_create_method(self):
        self.assertTrue(hasattr(self.model, "create"))
        self.assertTrue(callable(self.model.create))

    def test_create_without_parameters_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingParameterError):
            self.model.create()

    def test_create_without_variables_raises(self):
        self.model.parameters = [0]
        with self.assertRaises(aspecd.exceptions.MissingParameterError):
            self.model.create()

    def test_create_with_missing_parameter_raises(self):
        class MyModel(model.Model):
            def _sanitise_parameters(self):
                if "coefficient" not in self.parameters:
                    raise aspecd.exceptions.MissingParameterError(
                        message="Parameter 'coefficient' missing"
                    )

        model_ = MyModel()
        model_.parameters["foo"] = "bar"
        model_.variables = [np.linspace(0, 1)]
        with self.assertRaisesRegex(
            aspecd.exceptions.MissingParameterError, "coefficient"
        ):
            model_.create()

    def test_create_returns_calculated_dataset(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        self.assertIsInstance(dataset, aspecd.dataset.CalculatedDataset)

    def test_create_sets_calculated_dataset_axis_values(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        np.testing.assert_allclose(
            dataset.data.axes[0].values, self.model.variables[0]
        )

    def test_create_sets_last_axis_quantity_and_unit_to_defaults(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        self.assertEqual("amplitude", dataset.data.axes[-1].quantity)
        self.assertEqual("a.u.", dataset.data.axes[-1].unit)

    def test_create_with_axes_sets_axes_quantities_and_units(self):
        axes = [
            {"quantity": "foo0", "unit": "bar0"},
            {"quantity": "foo1", "unit": "bar1"},
        ]
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        self.model.axes = axes
        dataset = self.model.create()
        for idx, axis in enumerate(dataset.data.axes):
            self.assertEqual(axes[idx]["quantity"], axis.quantity)
            self.assertEqual(axes[idx]["unit"], axis.unit)

    def test_create_with_axes_and_omitted_axis(self):
        axes = [
            None,
            {"quantity": "foo1", "unit": "bar1"},
        ]
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        self.model.axes = axes
        dataset = self.model.create()
        for idx, axis in enumerate(dataset.data.axes):
            if axes[idx]:
                self.assertEqual(axes[idx]["quantity"], axis.quantity)
                self.assertEqual(axes[idx]["unit"], axis.unit)
            else:
                self.assertEqual("", axis.quantity)
                self.assertEqual("", axis.unit)

    def test_create_with_axes_unequal_to_number_of_dataset_axes_raises(self):
        axes = [{"quantity": "foo0", "unit": "bar0"}]
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        self.model.axes = axes
        message = "Number of axes and dataset axes need to be identical"
        with self.assertRaisesRegex(IndexError, message):
            self.model.create()

    def test_create_sets_calculated_dataset_origdata_axis_values(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        np.testing.assert_allclose(
            dataset._origdata.axes[0].values, self.model.variables[0]
        )

    def test_create_with_2d_sets_calculated_dataset_axis_values(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1), np.linspace(2, 3)]
        dataset = self.model.create()
        for index in range(len(self.model.variables)):
            np.testing.assert_allclose(
                dataset.data.axes[index].values, self.model.variables[index]
            )

    def test_create_with_2d_sets_last_axis_measure_and_unit_to_defaults(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1), np.linspace(2, 3)]
        dataset = self.model.create()
        self.assertEqual("amplitude", dataset.data.axes[-1].quantity)
        self.assertEqual("a.u.", dataset.data.axes[-1].unit)

    def test_create_sets_calculated_dataset_calculation_type(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        self.assertEqual(self.model.name, dataset.metadata.calculation.type)

    def test_create_sets_calculated_dataset_calculation_parameters(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        dataset = self.model.create()
        self.assertEqual(
            self.model.parameters, dataset.metadata.calculation.parameters
        )

    def test_create_with_label_sets_label_of_calculated_dataset(self):
        self.model.parameters = [0]
        self.model.variables = [np.linspace(0, 1)]
        self.model.label = "foobar"
        dataset = self.model.create()
        self.assertEqual(self.model.label, dataset.label)

    def test_has_from_dataset_method(self):
        self.assertTrue(hasattr(self.model, "from_dataset"))
        self.assertTrue(callable(self.model.from_dataset))

    def test_from_dataset_without_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.model.from_dataset()

    def test_from_dataset_sets_values(self):
        values = np.linspace(5, 50)
        dataset = aspecd.dataset.Dataset()
        dataset.data.data = np.linspace(0, 1)
        dataset.data.axes[0].values = values
        self.model.from_dataset(dataset=dataset)
        np.testing.assert_allclose(values, self.model.variables[0])

    def test_from_2d_dataset_sets_values(self):
        dataset = aspecd.dataset.Dataset()
        dataset.data.data = np.random.random([10, 5])
        self.model.from_dataset(dataset=dataset)
        for index in range(len(dataset.data.axes) - 1):
            np.testing.assert_allclose(
                dataset.data.axes[index].values, self.model.variables[index]
            )

    def test_from_dataset_applies_dataset_axes_to_calculated_dataset(self):
        dataset = aspecd.dataset.Dataset()
        dataset.data.data = np.linspace(0, 1)
        dataset.data.axes[0].quantity = "foo"
        dataset.data.axes[1].quantity = "bar"
        self.model.from_dataset(dataset=dataset)
        self.model.parameters = [0]
        calculated_dataset = self.model.create()
        for index in range(len(dataset.data.axes)):
            self.assertEqual(
                dataset.data.axes[index].quantity,
                calculated_dataset.data.axes[index].quantity,
            )

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.model, "from_dict"))
        self.assertTrue(callable(self.model.from_dict))

    def test_from_dict_without_dict_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDictError):
            self.model.from_dict()

    def test_from_dict_sets_parameters(self):
        dict_ = {"parameters": {"foo": 42}, "variables": [np.linspace(0, 1)]}
        self.model.from_dict(dict_)
        self.assertDictEqual(dict_["parameters"], self.model.parameters)

    def test_from_dict_sets_variables(self):
        dict_ = {"parameters": {"foo": 42}, "variables": [np.linspace(0, 1)]}
        self.model.from_dict(dict_)
        self.assertEqual(dict_["variables"], self.model.variables)

    def test_from_dict_sets_only_valid_properties(self):
        dict_ = {"foo": 42}
        self.model.from_dict(dict_)
        self.assertFalse(hasattr(self.model, "foo"))

    def test_has_evaluate_method(self):
        self.assertTrue(hasattr(self.model, "evaluate"))
        self.assertTrue(callable(self.model.evaluate))

    def test_evaluate_returns_data(self):
        self.model.parameters = {"foo": 42}
        self.model.variables = [np.linspace(0, 1)]
        data = self.model.evaluate()
        self.assertListEqual(list(self.model.variables[0]), list(data))


class TestCompositeModel(unittest.TestCase):
    def setUp(self):
        self.model = model.CompositeModel()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "composite model consisting of several weighted models",
            self.model.description.lower(),
        )

    def test_has_models_property(self):
        self.assertTrue(hasattr(self.model, "models"))

    def test_has_weights_property(self):
        self.assertTrue(hasattr(self.model, "weights"))

    def test_create_with_single_model_equivalent_to_single_model(self):
        variables = np.linspace(0, 5)
        models = ["Polynomial"]
        parameters = {"coefficients": [1]}
        # Create single model
        single_model = model.Polynomial()
        single_model.parameters = parameters
        single_model.variables = [variables]
        single_model_result = single_model.create()
        # Create composite model
        self.model.models = models
        self.model.parameters = [parameters]
        self.model.variables = [variables]
        composite_model_result = self.model.create()
        self.assertListEqual(
            list(single_model_result.data.data),
            list(composite_model_result.data.data),
        )

    def test_weighting_with_single_model(self):
        variables = np.linspace(0, 5)
        models = ["Polynomial"]
        parameters = {"coefficients": [1]}
        weights = [2]
        # Create single model
        single_model = model.Polynomial()
        single_model.parameters = parameters
        single_model.variables = [variables]
        single_model_result = single_model.create()
        # Create composite model
        self.model.models = models
        self.model.parameters = [parameters]
        self.model.weights = weights
        self.model.variables = [variables]
        composite_model_result = self.model.create()
        self.assertListEqual(
            list(single_model_result.data.data * weights[0]),
            list(composite_model_result.data.data),
        )

    def test_create_with_multiple_models_equivalent_to_sum_of_models(self):
        variables = [np.linspace(0, 5)]
        models = ["Sine", "Exponential"]
        parameters = [{"amplitude": 10}, {"rate": -4}]
        # Create individual models
        data = np.zeros(len(variables[0]))
        for idx, model_name in enumerate(models):
            model_ = aspecd.utils.object_from_class_name(
                "aspecd.model." + model_name
            )
            for key in parameters[idx]:
                # noinspection PyUnresolvedReferences
                model_.parameters[key] = parameters[idx][key]
            model_.variables = variables
            # noinspection PyUnresolvedReferences
            model_result = model_.create()
            data += model_result.data.data
        # Create composite model
        self.model.models = models
        self.model.parameters = parameters
        self.model.variables = variables
        composite_model_result = self.model.create()
        self.assertListEqual(
            list(data), list(composite_model_result.data.data)
        )

    def test_create_with_operator(self):
        variables = [np.linspace(0, 5)]
        models = ["Sine", "Exponential"]
        parameters = [{"amplitude": 10}, {"rate": -4}]
        operators = ["*"]
        # Create individual models
        data = np.zeros(len(variables[0]))
        for idx, model_name in enumerate(models):
            model_ = aspecd.utils.object_from_class_name(
                "aspecd.model." + model_name
            )
            for key in parameters[idx]:
                # noinspection PyUnresolvedReferences
                model_.parameters[key] = parameters[idx][key]
            model_.variables = variables
            # noinspection PyUnresolvedReferences
            model_result = model_.create()
            if not idx:
                data += model_result.data.data
            else:
                data *= model_result.data.data
        # Create composite model
        self.model.models = models
        self.model.parameters = parameters
        self.model.variables = variables
        self.model.operators = operators
        composite_model_result = self.model.create()
        self.assertListEqual(
            list(data), list(composite_model_result.data.data)
        )

    def test_create_with_incompatible_no_of_models_and_parameters_raises(
        self,
    ):
        self.model.models = ["Sine", "Exponential"]
        self.model.parameters = [{"amplitude": 10}]
        self.model.variables = np.linspace(0, 5)
        with self.assertRaisesRegex(
            IndexError, "Models and parameters count differs"
        ):
            self.model.create()

    def test_create_with_incompatible_no_of_weights(self):
        self.model.models = ["Sine", "Exponential"]
        self.model.parameters = [{"amplitude": 10}, {"rate": -4}]
        self.model.weights = [2]
        self.model.variables = np.linspace(0, 5)
        with self.assertRaisesRegex(
            IndexError, "Models and weights count differs"
        ):
            self.model.create()

    def test_create_with_incompatible_no_of_operators(self):
        self.model.models = ["Sine", "Exponential", "Sine"]
        self.model.parameters = [
            {"amplitude": 10},
            {"rate": -4},
            {"frequency": 0.5},
        ]
        self.model.operators = ["*"]
        self.model.variables = np.linspace(0, 5)
        with self.assertRaisesRegex(
            IndexError, "Models and operators count differs"
        ):
            self.model.create()


class TestFamilyOfCurves(unittest.TestCase):
    def setUp(self):
        self.model = model.FamilyOfCurves()
        self.variables = np.linspace(0, 6 * np.pi)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "family of curves for a model with one parameter varied",
            self.model.description.lower(),
        )

    def test_has_model_property(self):
        self.assertTrue(hasattr(self.model, "model"))

    def test_has_vary_property(self):
        self.assertTrue(hasattr(self.model, "vary"))

    def test_create_without_model_raises(self):
        with self.assertRaisesRegex(ValueError, "Missing a model"):
            self.model.create()

    def test_create_model_with_scalar_varied_parameter_value(self):
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = 4
        self.model.variables = [self.variables]
        simple_model = model.Sine()
        simple_model.variables = [self.variables]
        simple_model.parameters["amplitude"] = self.model.vary["values"]
        simple_dataset = simple_model.create()
        family_of_curves = self.model.create()
        self.assertListEqual(
            list(simple_dataset.data.data), list(family_of_curves.data.data)
        )

    def test_create_model_with_actually_varied_parameter_value(self):
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2, 4]
        self.model.variables = [self.variables]
        simple_model1 = model.Sine()
        simple_model1.variables = [self.variables]
        simple_model1.parameters["amplitude"] = self.model.vary["values"][0]
        simple_dataset1 = simple_model1.create()
        simple_model2 = copy.deepcopy(simple_model1)
        simple_model2.parameters["amplitude"] = self.model.vary["values"][1]
        simple_dataset2 = simple_model2.create()
        family_of_curves = self.model.create()
        self.assertListEqual(
            list(simple_dataset1.data.data),
            list(family_of_curves.data.data[:, 0]),
        )
        self.assertListEqual(
            list(simple_dataset2.data.data),
            list(family_of_curves.data.data[:, 1]),
        )

    def test_create_model_sets_quantity_of_additional_axis(self):
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2, 4]
        self.model.variables = [self.variables]
        family_of_curves = self.model.create()
        self.assertEqual(
            self.model.vary["parameter"],
            family_of_curves.data.axes[-2].quantity,
        )

    def test_create_model_sets_axis_values(self):
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2, 4]
        self.model.variables = [self.variables]
        family_of_curves = self.model.create()
        self.assertListEqual(
            self.model.vary["values"],
            list(family_of_curves.data.axes[-2].values),
        )

    def test_create_model_sets_float_axis_values(self):
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2.0, 4.0]
        self.model.variables = [self.variables]
        family_of_curves = self.model.create()
        self.assertListEqual(
            self.model.vary["values"],
            list(family_of_curves.data.axes[-2].values),
        )

    def test_create_from_dataset(self):
        values = np.linspace(5, 50)
        dataset = aspecd.dataset.Dataset()
        dataset.data.data = np.linspace(0, 1)
        dataset.data.axes[0].values = values
        self.model.from_dataset(dataset=dataset)
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2, 4]
        family_of_curves = self.model.create()
        self.assertEqual(
            self.model.vary["parameter"],
            family_of_curves.data.axes[-2].quantity,
        )

    def test_create_with_axes_unequal_to_number_of_dataset_axes_raises(self):
        axes = [
            {"quantity": "foo0", "unit": "bar0"},
            {"quantity": "foo0", "unit": "bar0"},
        ]
        values = np.linspace(5, 50)
        dataset = aspecd.dataset.Dataset()
        dataset.data.data = np.linspace(0, 1)
        dataset.data.axes[0].values = values
        self.model.from_dataset(dataset=dataset)
        self.model.model = "Sine"
        self.model.vary["parameter"] = "amplitude"
        self.model.vary["values"] = [2, 4]
        self.model.axes = axes
        message = "Number of axes and dataset axes need to be identical"
        with self.assertRaisesRegex(IndexError, message):
            self.model.create()


class TestZeros(unittest.TestCase):
    def setUp(self):
        self.model = model.Zeros()
        self.dataset = aspecd.dataset.CalculatedDataset()
        self.dataset.data.data = np.random.randn(10)
        self.dataset3d = aspecd.dataset.CalculatedDataset()
        self.dataset3d.data.data = np.random.randn(10, 10, 10)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("containing only zeros", self.model.description.lower())

    def test_create_without_coefficients_raises(self):
        with self.assertRaisesRegex(
            aspecd.exceptions.MissingParameterError, "shape"
        ):
            self.model.create()

    def test_create_returns_zeros_of_given_shape(self):
        self.model.parameters["shape"] = 5
        dataset = self.model.create()
        self.assertListEqual(list(np.zeros(5)), list(dataset.data.data))

    def test_create_returns_zeros_of_given_3d_shape(self):
        self.model.parameters["shape"] = [5, 5, 5]
        dataset = self.model.create()
        self.assertFalse(all(dataset.data.data.flatten()))
        self.assertEqual((5, 5, 5), dataset.data.data.shape)

    def test_create_from_1d_dataset_returns_zeros_with_correct_shape(self):
        self.model.from_dataset(self.dataset)
        dataset = self.model.create()
        self.assertEqual(len(self.dataset.data.data), len(dataset.data.data))

    def test_create_from_3d_dataset_returns_zeros_with_correct_shape(self):
        self.model.from_dataset(self.dataset3d)
        dataset = self.model.create()
        self.assertEqual(
            len(self.dataset3d.data.data), len(dataset.data.data)
        )

    def test_create_1d_with_range_sets_axis_range(self):
        self.model.parameters["shape"] = 5
        self.model.parameters["range"] = [35, 42]
        dataset = self.model.create()
        self.assertEqual(
            self.model.parameters["range"][0][0],
            dataset.data.axes[0].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][0][-1],
            dataset.data.axes[0].values[-1],
        )

    def test_create_2d_with_range_sets_axis_range(self):
        self.model.parameters["shape"] = [5, 5]
        self.model.parameters["range"] = [[35, 42], [17.5, 21]]
        dataset = self.model.create()
        self.assertEqual(
            self.model.parameters["range"][0][0],
            dataset.data.axes[0].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][0][-1],
            dataset.data.axes[0].values[-1],
        )
        self.assertEqual(
            self.model.parameters["range"][1][0],
            dataset.data.axes[1].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][1][-1],
            dataset.data.axes[1].values[-1],
        )

    def test_range_incompatible_to_shape_raises(self):
        self.model.parameters["shape"] = 5
        self.model.parameters["range"] = [[35, 42], [17.5, 21]]
        with self.assertRaisesRegex(
            IndexError, "Shape and range must be compatible"
        ):
            self.model.create()


class TestOnes(unittest.TestCase):
    def setUp(self):
        self.model = model.Ones()
        self.dataset = aspecd.dataset.CalculatedDataset()
        self.dataset.data.data = np.random.randn(10)
        self.dataset3d = aspecd.dataset.CalculatedDataset()
        self.dataset3d.data.data = np.random.randn(10, 10, 10)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("containing only ones", self.model.description.lower())

    def test_create_without_coefficients_raises(self):
        with self.assertRaisesRegex(
            aspecd.exceptions.MissingParameterError, "shape"
        ):
            self.model.create()

    def test_create_returns_ones_of_given_shape(self):
        self.model.parameters["shape"] = 5
        dataset = self.model.create()
        self.assertListEqual(list(np.ones(5)), list(dataset.data.data))

    def test_create_returns_ones_of_given_3d_shape(self):
        self.model.parameters["shape"] = [5, 5, 5]
        dataset = self.model.create()
        self.assertFalse(all(dataset.data.data.flatten() - 1))
        self.assertEqual((5, 5, 5), dataset.data.data.shape)

    def test_create_from_1d_dataset_returns_zeros_with_correct_shape(self):
        self.model.from_dataset(self.dataset)
        dataset = self.model.create()
        self.assertEqual(len(self.dataset.data.data), len(dataset.data.data))

    def test_create_from_3d_dataset_returns_zeros_with_correct_shape(self):
        self.model.from_dataset(self.dataset3d)
        dataset = self.model.create()
        self.assertEqual(
            len(self.dataset3d.data.data), len(dataset.data.data)
        )

    def test_create_1d_with_range_sets_axis_range(self):
        self.model.parameters["shape"] = 5
        self.model.parameters["range"] = [35, 42]
        dataset = self.model.create()
        self.assertEqual(
            self.model.parameters["range"][0][0],
            dataset.data.axes[0].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][0][-1],
            dataset.data.axes[0].values[-1],
        )

    def test_create_2d_with_range_sets_axis_range(self):
        self.model.parameters["shape"] = [5, 5]
        self.model.parameters["range"] = [[35, 42], [17.5, 21]]
        dataset = self.model.create()
        self.assertEqual(
            self.model.parameters["range"][0][0],
            dataset.data.axes[0].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][0][-1],
            dataset.data.axes[0].values[-1],
        )
        self.assertEqual(
            self.model.parameters["range"][1][0],
            dataset.data.axes[1].values[0],
        )
        self.assertEqual(
            self.model.parameters["range"][1][-1],
            dataset.data.axes[1].values[-1],
        )

    def test_range_incompatible_to_shape_raises(self):
        self.model.parameters["shape"] = 5
        self.model.parameters["range"] = [[35, 42], [17.5, 21]]
        with self.assertRaisesRegex(
            IndexError, "Shape and range must be compatible"
        ):
            self.model.create()


class TestPolynomial(unittest.TestCase):
    def setUp(self):
        self.model = model.Polynomial()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("polynomial", self.model.description.lower())

    def test_create_returns_polynomial(self):
        self.model.parameters["coefficients"] = [0, 1]
        self.model.variables = [np.linspace(0, 5, 6)]
        dataset = self.model.create()
        self.assertEqual(0, dataset.data.data[0])
        self.assertEqual(5, dataset.data.data[-1])

    def test_create_without_coefficients_raises(self):
        self.model.variables = np.linspace(0, 5, 6)
        with self.assertRaisesRegex(
            aspecd.exceptions.MissingParameterError, "coefficient"
        ):
            self.model.create()


def get_fwhm(dataset):
    xdata = dataset.data.axes[0].values
    ydata = dataset.data.data
    max_value = np.max(ydata)
    max_pos = np.argmax(ydata)
    halfmax_pos_left = np.argmin(abs(ydata[:max_pos] - (max_value / 2)))
    halfmax_pos_right = (
        np.argmin(abs(ydata[max_pos:] - (max_value / 2))) + max_pos
    )
    fwhm = xdata[halfmax_pos_right] - xdata[halfmax_pos_left]
    return fwhm


class TestGaussian(unittest.TestCase):
    def setUp(self):
        self.model = model.Gaussian()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("generalised gaussian", self.model.description.lower())

    def test_amplitude_sets_maximum(self):
        amplitude = 0.5
        self.model.variables = [np.linspace(-5, 5, 2**10)]
        self.model.parameters["amplitude"] = amplitude
        dataset = self.model.create()
        self.assertAlmostEqual(amplitude, max(dataset.data.data), 4)

    def test_position_set_position(self):
        position = 0.5
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["position"] = position
        dataset = self.model.create()
        self.assertAlmostEqual(
            position, self.model.variables[0][np.argmax(dataset.data.data)], 4
        )

    def test_width_sets_width(self):
        width = 2
        self.model.variables = [np.linspace(-5, 5, 10001)]
        self.model.parameters["width"] = width
        dataset = self.model.create()
        fwhm_expected = 2 * np.sqrt(2 * np.log(2)) * width
        self.assertAlmostEqual(fwhm_expected, get_fwhm(dataset), 3)

    def test_zero_width_returns_finite_output(self):
        width = 0
        self.model.variables = [np.linspace(-5, 5)]
        self.model.parameters["width"] = width
        with np.errstate(divide="raise"):
            dataset = self.model.create()
        self.assertTrue(np.all(np.isfinite(dataset.data.data)))


class TestNormalisedGaussian(unittest.TestCase):
    def setUp(self):
        self.model = model.NormalisedGaussian()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("normalised gaussian", self.model.description.lower())

    def test_amplitude_is_normalised(self):
        self.model.variables = [np.linspace(-5, 5, 2**10)]
        amplitude = 1 / (self.model.parameters["width"] * np.sqrt(2 * np.pi))
        dataset = self.model.create()
        self.assertAlmostEqual(amplitude, max(dataset.data.data), 4)

    def test_position_set_position(self):
        position = 0.5
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["position"] = position
        dataset = self.model.create()
        self.assertAlmostEqual(
            position, self.model.variables[0][np.argmax(dataset.data.data)], 4
        )

    def test_width_sets_width(self):
        width = 2
        self.model.variables = [np.linspace(-5, 5, 10001)]
        self.model.parameters["width"] = width
        dataset = self.model.create()
        fwhm_expected = 2 * np.sqrt(2 * np.log(2)) * width
        self.assertAlmostEqual(fwhm_expected, get_fwhm(dataset), 3)

    def test_zero_width_returns_finite_output(self):
        width = 0
        self.model.variables = [np.linspace(-5, 5)]
        self.model.parameters["width"] = width
        with np.errstate(divide="raise"):
            dataset = self.model.create()
        self.assertTrue(np.all(np.isfinite(dataset.data.data)))


class TestLorentzian(unittest.TestCase):
    def setUp(self):
        self.model = model.Lorentzian()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "generalised lorentzian", self.model.description.lower()
        )

    def test_amplitude_sets_maximum(self):
        amplitude = 0.5
        self.model.variables = [np.linspace(-5, 5, 2**10)]
        self.model.parameters["amplitude"] = amplitude
        dataset = self.model.create()
        self.assertAlmostEqual(amplitude, max(dataset.data.data), 4)

    def test_position_set_position(self):
        position = 0.5
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["position"] = position
        dataset = self.model.create()
        self.assertAlmostEqual(
            position, self.model.variables[0][np.argmax(dataset.data.data)], 4
        )

    def test_width_sets_width(self):
        width = 2
        self.model.variables = [np.linspace(-5, 5, 10001)]
        self.model.parameters["width"] = width
        dataset = self.model.create()
        fwhm_expected = 2 * width
        self.assertAlmostEqual(fwhm_expected, get_fwhm(dataset), 3)

    def test_zero_width_returns_finite_output(self):
        width = 0
        self.model.variables = [np.linspace(-5, 5)]
        self.model.parameters["width"] = width
        with np.errstate(divide="raise"):
            dataset = self.model.create()
        self.assertTrue(np.all(np.isfinite(dataset.data.data)))


class TestNormalisedLorentzian(unittest.TestCase):
    def setUp(self):
        self.model = model.NormalisedLorentzian()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("normalised lorentzian", self.model.description.lower())

    def test_amplitude_is_normalised(self):
        self.model.variables = [np.linspace(-5, 5, 2**10)]
        amplitude = 1 / (self.model.parameters["width"] * np.pi)
        dataset = self.model.create()
        self.assertAlmostEqual(amplitude, max(dataset.data.data), 4)

    def test_position_set_position(self):
        position = 0.5
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["position"] = position
        dataset = self.model.create()
        self.assertAlmostEqual(
            position, self.model.variables[0][np.argmax(dataset.data.data)], 4
        )

    def test_width_sets_width(self):
        width = 2
        self.model.variables = [np.linspace(-5, 5, 10001)]
        self.model.parameters["width"] = width
        dataset = self.model.create()
        fwhm_expected = 2 * width
        self.assertAlmostEqual(fwhm_expected, get_fwhm(dataset), 3)

    def test_zero_width_returns_finite_output(self):
        width = 0
        self.model.variables = [np.linspace(-5, 5)]
        self.model.parameters["width"] = width
        with np.errstate(divide="raise"):
            dataset = self.model.create()
        self.assertTrue(np.all(np.isfinite(dataset.data.data)))


class TestVoigtian(unittest.TestCase):
    def setUp(self):
        self.model = model.Voigtian()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("voigt profile", self.model.description.lower())

    def test_default_position_is_zero(self):
        self.model.variables = [np.linspace(-5, 5, 1001)]
        dataset = self.model.create()
        self.assertEqual(
            0, self.model.variables[0][dataset.data.data.argmax()]
        )

    def test_position_sets_position(self):
        position = 0.5
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["position"] = position
        dataset = self.model.create()
        self.assertAlmostEqual(
            position, self.model.variables[0][dataset.data.data.argmax()], 4
        )

    def test_gamma_zero_equals_normalised_gaussian(self):
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["gamma"] = 0
        dataset = self.model.create()
        gaussian = model.NormalisedGaussian()
        gaussian.variables = self.model.variables
        gaussian_data = gaussian.create()
        np.testing.assert_allclose(dataset.data.data, gaussian_data.data.data)

    def test_sigma_zero_equals_normalised_lorentzian(self):
        self.model.variables = [np.linspace(-5, 5, 1001)]
        self.model.parameters["sigma"] = 0
        dataset = self.model.create()
        lorentzian = model.NormalisedLorentzian()
        lorentzian.variables = self.model.variables
        lorentzian_data = lorentzian.create()
        np.testing.assert_allclose(
            dataset.data.data, lorentzian_data.data.data
        )


class TestSine(unittest.TestCase):
    def setUp(self):
        self.model = model.Sine()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("sine", self.model.description.lower())

    def test_roots_at_correct_place(self):
        self.model.variables = [np.linspace(0, 2 * np.pi, 1001)]
        dataset = self.model.create()
        for index in [0, 500, 1000]:
            self.assertAlmostEqual(0, dataset.data.data[index])
        self.assertTrue(np.all(dataset.data.data[1:499]))
        self.assertTrue(np.all(dataset.data.data[501:1000]))

    def test_amplitude_sets_amplitude(self):
        amplitude = 42.0
        self.model.variables = [np.linspace(0, 2 * np.pi, 1001)]
        self.model.parameters["amplitude"] = amplitude
        dataset = self.model.create()
        self.assertAlmostEqual(amplitude, max(dataset.data.data))

    def test_frequency_sets_frequency(self):
        self.model.variables = [np.linspace(0, 2 * np.pi, 1001)]
        self.model.parameters["frequency"] = 2
        dataset = self.model.create()
        for index in [0, 250, 500, 750, 1000]:
            self.assertAlmostEqual(0, dataset.data.data[index])

    def test_phase_shifts_roots(self):
        self.model.variables = [np.linspace(0, 2 * np.pi, 1001)]
        self.model.parameters["phase"] = 0.5 * np.pi
        dataset = self.model.create()
        for index in [250, 750]:
            self.assertAlmostEqual(0, dataset.data.data[index])


class TestExponential(unittest.TestCase):
    def setUp(self):
        self.model = model.Exponential()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("exponential", self.model.description.lower())

    def test_only_positive_values(self):
        self.model.variables = [np.linspace(-1, 2, 101)]
        dataset = self.model.create()
        self.assertTrue(np.all(dataset.data.data > 0))

    def test_prefactor_sets_intercept(self):
        prefactor = 42.0
        self.model.variables = [np.linspace(0, 2, 101)]
        self.model.parameters["prefactor"] = prefactor
        dataset = self.model.create()
        self.assertAlmostEqual(prefactor, dataset.data.data[0])

    def test_rate_sets_rate(self):
        rate = 4.2
        self.model.variables = [np.linspace(0, 2, 101)]
        self.model.parameters["rate"] = rate
        dataset = self.model.create()
        self.assertAlmostEqual(
            np.exp(rate * self.model.variables[0][-1]), dataset.data.data[-1]
        )
